/* $Id: etunbridge.c,v 1.23 2001/10/12 15:19:40 hiddy Exp $ */

/*-
 * Copyright (c) 2001 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/sockio.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define DEFAULT_TUN	"etun0"
#define DEFAULT_PORT	12345

/* ioctl's for get/set debug */
#define ETUNIOSDEBUG	_IOW('e', 100, int)
#define ETUNIOSHASFCS	_IOW('e', 101, int)
#define ETUNIOIFATTACH	_IOW('e', 102, struct ifreq *)
#define ETUNIOIFDETACH	_IOR('e', 103, int)

int readn (int fd, char *ptr, int nbytes);
int writen(int fd, char *ptr, int nbytes);
int open_tundev(char *, char *);
int ifsetflags(int, char *, int flag, int on);
#define ON 1
#define OFF 0

int socket_reuse(int sockfd, int yes);
int tcp_bridge(int soc, int tunfd);
int udp_bridge(int soc, int tunfd);

int 	soc = -1, 	/* socket */
	tunfd = -1;     /* ethertunnel */
char 	*tunif = DEFAULT_TUN, 
	*bridgeif = NULL;

int ifenable(void);
int ifdisable(void);

void endall(int sig)
{
	struct ifreq ifr;

	if(soc >= 0){
		memcpy(ifr.ifr_name, tunif, IFNAMSIZ);
		ioctl(soc, SIOCGIFFLAGS, &ifr);
                ifr.ifr_flags &= ~IFF_UP;
        	ioctl(soc, SIOCSIFFLAGS, &ifr);

		if(bridgeif){
			memcpy(ifr.ifr_name, tunif, IFNAMSIZ);
			ioctl(soc, SIOCGIFFLAGS, &ifr);
			ifr.ifr_flags &= ~(IFF_UP);
			ioctl(soc, SIOCSIFFLAGS, &ifr);
		}
		close(soc);
	}
	if(tunfd >= 0)
		close(tunfd);
	soc = tunfd = -1;
}

void
usage_exit()
{
	fprintf(stderr, 
	  "Usage: etunbridge [-dstuv] <-p port> <-i tunif> dstipaddr\n");
	fprintf(stderr, 
	  "\t-d:		enable etun debug mode\n"
	  "\t-s 		server mode(TCP:listen, UDP:passive)\n"
	  "\t-t:		disable stripping FCS trailer\n"
	  "\t-u 		use UDP(default TCP)\n"
	  "\t-v:		enable verbose mode\n"
	  "\t-i etunif:	ethernet tunnel ifname(default: etun0)\n"
	  "\t-b ethif:	bridge ethernet ifname\n"
	  "\t-p port:	dst port number(default: %d)\n" 
	  "\t-P port:	src port number(default: %d)\n" 		
	   ,DEFAULT_PORT,DEFAULT_PORT);
		
	exit(1);
}

#define VPRINTF(x)	if(vopt) 	printf x
int vopt = 0, 	/* verbose mode */
    dopt = 0, 	/* etun debug mode */
    topt = 0,	/* disabel stripping FCS trailer */
    sopt = 0, 	/* server mode */
    uopt = 0;	/* use UDP */

int 
main(int argc, char **argv)
{
	char *hostname, *srcaddr = NULL;
	int c, dport = DEFAULT_PORT, sport = DEFAULT_PORT;
	struct sockaddr_in sin;
	struct hostent *hp;
	int (*bridge)(int, int) = tcp_bridge;
	
	while((c = getopt(argc, argv, "i:b:p:P:S:dstuv")) != EOF){
		switch(c){
		case 'd':
			dopt++;
			break;
		case 's':
			sopt++;
			break;
		case 't':
			topt++;
			break;
		case 'u':
			uopt++;
			break;
		case 'v':
			vopt++;
			break;
		case 'i': 
			tunif = optarg;
			break;
		case 'b':
			bridgeif = optarg;
			break;
		case 'P':
			sport = atoi(optarg);
			break;
		case 'p':
			dport = atoi(optarg);
			break;
		case 'S':
			srcaddr = optarg;
			break;
		default:
			usage_exit();
		}
	}

	argc -= optind;
	argv += optind;

	if(argc != 1)
		usage_exit(1);	

	VPRINTF(("- options\n"));
	VPRINTF(("\tetun debug mode:\t\t%s\n", dopt ? "enable": "disable"));
	VPRINTF(("\tverbose mode:   \t\t%s\n", vopt ? "enable": "disable"));
	VPRINTF(("\tserver:         \t\t%s\n", sopt ? "true": "false"));
	VPRINTF(("\tuse UDP:        \t\t%s\n", uopt ? "true": "false"));
	VPRINTF(("\tstrip FCS:      \t\t%s\n", !topt ? "enable": "disable"));
	VPRINTF(("\tdst port:       \t\t%d\n", dport));
	VPRINTF(("\tsrc port:       \t\t%d\n", sport));
	VPRINTF(("\tethernet tunif: \t\t%s\n", tunif));

	signal(SIGINT, endall);
	atexit((void (*)(void))endall);

	if(uopt)
		bridge = udp_bridge;

	/* open ethernet tunnel device */
	tunfd = open_tundev(tunif, bridgeif);
	if(tunfd < 0){
		perror("open_tundev()");
		exit(1);
	}

	/* open socket */
	if(( soc = socket(AF_INET, uopt? SOCK_DGRAM: SOCK_STREAM, 0)) < 0){
		perror("socket");
	        exit(1);
	}
	VPRINTF(("etunbridge: socket open\n"));

	/*  udp or TCP server mode: bind local address */
	if(uopt || sopt){  
		memset(&sin, 0, sizeof(sin));
		socket_reuse(sopt, 1);
		sin.sin_len = sizeof(sin);
		sin.sin_family = AF_INET;
		sin.sin_port = htons(sport);
		if(srcaddr != NULL)
			sin.sin_addr.s_addr = inet_addr(srcaddr);
		else
			sin.sin_addr.s_addr = 0;
		if(bind(soc, (struct sockaddr *)&sin, sizeof(sin)) < 0){
			perror("bind");
			exit(1);
		}
		VPRINTF(("etunbridge: bind\n"));
	}

	/* resolve name */
	hostname = argv[0];
	if((hp = gethostbyname(hostname)) == NULL){
		fprintf(stderr, "%s: unknown host.\n", hostname);
	       	exit(1);
	}
  
	/* UDP || TCP client */
	if(uopt || !sopt){
		memset(&sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_port = htons(dport);
		memcpy(&sin.sin_addr, hp->h_addr, hp->h_length);
		if(connect(soc, (struct sockaddr *)&sin, sizeof(sin)) < 0){
			perror("connect");
			exit(1);
		}
		VPRINTF(("etunbridge: connect\n"));
	}

	/* UDP or TCP client */
	if(uopt || !sopt){
		/* UDP passive mode: wait for packet */
		if(sopt){
			int nfds = soc + 1; 
			fd_set fds;
			FD_ZERO(&fds);
			FD_SET(soc, &fds);

			VPRINTF(("etunbridge: waiting for packet from peer ...\n"));
			select(nfds, &fds, NULL, NULL, NULL);
		}

		ifenable();
		(void)bridge(soc, tunfd);
	}
	else {
		int csoc, len;
		listen(soc, 3);
		while(1){
			csoc = accept(soc, (struct sockaddr *)&sin, &len);
			if(csoc < 0){
				perror("accept");
				break;
			}
			if(memcmp(&sin.sin_addr, hp->h_addr, hp->h_length) != 0){
				close(csoc);
				continue;
			}
	
			ifenable();
			if(bridge(csoc, tunfd) < 0){
				perror("bridge()");
				break;
			}
			close(csoc);
			ifdisable();
		}
	}

	return 0;
}

int readn (int fd, char *ptr, int nbytes)
{
	int nleft;
	int nread;
	nleft = nbytes;

	while (nleft > 0){
		nread = read (fd, ptr, nleft);

		if (nread < 0) 
			return (nread);
		else if (nread == 0) 
			break;

		nleft -= nread;
		ptr += nread;
	}
	return nbytes - nleft;
}

int
writen(int fd, char *ptr, int nbytes)
{
	int nleft;
	int nwritten;

	nleft = nbytes;

	while (nleft > 0){
		nwritten = write(fd, ptr, nleft);
		
		if (nwritten <= 0) 
			return (nwritten);

		nleft -= nwritten;
		ptr += nwritten;
	}
	return nbytes - nleft;
}


int open_tundev(char *tunif, char *bridgeif)
{
	char dev[256];
	int t, ret, yes = 1, opt;
	struct ifreq ifr;
	strcpy(dev, "/dev/");
	strcat(dev, tunif);

	t = ret = open(dev, O_RDWR);
	if(t < 0)
		return ret;

	VPRINTF(("etunbridge: open(%s)\n", dev));
	
	if(dopt){
		if((ret = ioctl(t, ETUNIOSDEBUG, &yes)) < 0){
			perror("ioctl(ETUNIOSDEBUG)");
			return ret;
		}
		VPRINTF(("etunbridge: enable debug\n"));
	}

	opt = !topt;
	if((ret = ioctl(t, ETUNIOSHASFCS, &opt)) < 0){
		perror("ioctl(ETUNIOSHASFCS)");
		return ret;
	}
	VPRINTF(("etunbridge: %s strip fcs tailer\n", opt? "enable":"disable"));

  	if(bridgeif){
		memcpy(ifr.ifr_name, bridgeif, IFNAMSIZ);
		if((ret = ioctl(t, ETUNIOIFATTACH, &ifr)) < 0){
			perror("ioctl(ETUNIOIFATTACH)");
			return ret;
		}
		VPRINTF(("etunbridge: attach %s\n", bridgeif));
	}

	return t;
}

int ifenable(void){
	if(ifsetflags(soc, tunif, IFF_UP, ON) < 0)
		return -1;
	if(bridgeif && 
	   ifsetflags(soc, bridgeif, IFF_UP, ON) < 0)
		return -1;
	VPRINTF(("etunbridge: ifenable\n"));
	return 0;
}

int ifdisable(void){
	if(ifsetflags(soc, tunif, IFF_UP, OFF) < 0)
		return -1;
	if(bridgeif && 
	   ifsetflags(soc, bridgeif, IFF_UP, OFF) < 0)
		return -1;
	VPRINTF(("etunbridge: ifdisable\n"));
	return 0;
}

int ifsetflags(int s, char *ifname, int flags, int on)
{
	struct ifreq ifr;

	memcpy(ifr.ifr_name, ifname, IFNAMSIZ);
	if(ioctl(s, SIOCGIFFLAGS, &ifr) < 0){
		perror("ioctl(SIOCGIFFLAGS)");
		return  -1;
	}

	if(on)
		ifr.ifr_flags |= flags;
	else
		ifr.ifr_flags &= ~flags;

	if(ioctl(s, SIOCSIFFLAGS, &ifr) < 0){
		perror("ioctl(SIOCSIFFLAGS)");
		return -1;
	}
	VPRINTF(("etunbridge: interface %s flags: 0x%x \n", ifname, flags, on? "on": "off"));
	return 0;
}

int
socket_reuse (int s, int yes)
{
	int ret;

	if ((ret = setsockopt (s, SOL_SOCKET, SO_REUSEADDR, 
		(char *) &yes, sizeof (yes))) < 0) 
		return ret;
#ifdef SO_REUSEPORT
	ret = setsockopt (s, SOL_SOCKET, SO_REUSEPORT, 
		(char *) &yes, sizeof (yes));
#endif /* SO_REUSEPORT */
	return (ret);
}

int tcp_bridge(int s, int t)
{
	int nr, nw, nfds;
	char buf[1600];
	fd_set fds, rfds;
	u_long  pktlen;

	FD_ZERO(&fds);
	FD_SET(s, &fds);
	FD_SET(t, &fds);
	nfds = s + 1;

	while(1){
		rfds = fds;
		select(nfds, &rfds, NULL, NULL, NULL);
		if(FD_ISSET(t, &rfds)){
			nr = read(t, &buf[4], 1600);
			if(nr <= 0){
				perror("read(tun)");
				return nr;
			}
			VPRINTF(("etunbridge: read from tun %d bytes\n", nr));
			pktlen = htonl(nr);
			memcpy(buf, (char *)&pktlen, 4);
			nw = writen(s, buf, nr + 4);
			if(nw != nr + 4){
				perror("writen(2)");
				return nw;
			}
			VPRINTF(("etunbridge: send %d bytes\n", nw));
		}
		if(FD_ISSET(s, &rfds)){
			nr = readn(s, (char *)&pktlen, sizeof(pktlen));
			if(nr <= 0){
				perror("readn(socket) for hdr len");
				return nr;
			}
			VPRINTF(("etunbridge: recv hdr %d bytes\n", nr));
			pktlen = ntohl(pktlen);

			nr = readn(s, buf, pktlen);
			if(nr != pktlen){
				perror("readn(socket) for data");
				return nr;
			}
			VPRINTF(("etunbridge: recv %d bytes\n", nr));
		
			nw = writen(t, buf, pktlen);
			if(nw != pktlen){
				perror("write(tun)");
				return nw;
			}
			VPRINTF(("etunbridge: write to tun %d bytes\n", nw));
		}
	} 
	return -1;
}

int udp_bridge(int s, int t)
{
	int nr, nw, nfds;
	char buf[1600];
	fd_set fds, rfds;
	u_long  pktlen;

	FD_ZERO(&fds);
	FD_SET(s, &fds);
	FD_SET(t, &fds);
	nfds = s + 1;

	while(1){
		rfds = fds;
		select(nfds, &rfds, NULL, NULL, NULL);
		if(FD_ISSET(t, &rfds)){
			nr = read(t, &buf[4], 1600);
			if(nr <= 0){
				perror("read(tun)");
				return nr;
			}
			VPRINTF(("etunbridge: read from tun %d bytes\n", nr));
			pktlen = htonl(nr);
			memcpy(buf, (char *)&pktlen, 4);
			nw = writen(s, buf, nr + 4);
			if(nw != nr + 4){
				perror("writen(2)");
				return nw;
			}
			VPRINTF(("etunbridge: send %d bytes\n", nw));
		}
		if(FD_ISSET(s, &rfds)){
			nr = read(s, buf, 1600);
			if(nr <= 0){
				perror("readn(socket) for hdr len");
				return nr;
			}
			VPRINTF(("etunbridge: recv %d bytes\n", nr));
		
			memcpy((char *)&pktlen, buf, 4);
			pktlen = ntohl(pktlen);

			nw = writen(t, buf + 4, pktlen);
			if(nw != pktlen){
				perror("write(tun)");
				return nw;
			}
			VPRINTF(("etunbridge: write to tun %d bytes\n", nw));
		}
	} 
	return -1;
}
