/* $Id: if_etun.c,v 1.9 2001/10/12 15:19:40 hiddy Exp $ */

/*-
 * Copyright (c) 2001 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright (c) 1988, Julian Onions <jpo@cs.nott.ac.uk>
 * Nottingham University 1987.
 *
 * This source may be freely distributed, however I would be interested
 * in any changes that are made.
 *
 * This driver takes packets off the IP i/f and hands them up to a
 * user process to have its wicked way with. This driver has its
 * roots in a similar driver written by Phil Cockcroft (formerly) at
 * UCL. This driver is based much more on read/write/poll mode of
 * operation though.
 */

#include "etun.h"

#if NETUN > 0

#include "opt_inet.h"
#include "opt_ns.h"

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/select.h>
#include <sys/poll.h>
#include <sys/file.h>
#include <sys/signalvar.h>
#include <sys/conf.h>

#include <machine/cpu.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/if_ether.h>
#include <net/netisr.h>
#include <net/route.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_inarp.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#include "bpfilter.h"
#if NBPFILTER > 0
#include <sys/time.h>
#include <net/bpf.h>
#endif

struct etun_softc {
	u_short	etun_flags;		/* misc flags */
        struct  ethercom etun_ethercom;	/* ethernet common */
	struct	selinfo	etun_rsel;	/* read select */
	struct	selinfo	etun_wsel;	/* write select (not used) */
	int	etun_pgrp;		/* the process group - if any */
	struct  ifnet *bridgeif;	/* bridge ethernet if */
	struct  sockaddr_dl bifaddr;	/* save address */
};

/* etun_flags */
#define	ETUNF_OPEN	0x0001
#define	ETUNF_READY	ETUNF_OPEN
#define	ETUNF_RCOLL	0x0002
#define	ETUNF_RWAIT	0x0004
#define	ETUNF_ASYNC	0x0008
#define	ETUNF_NBIO	0x0010

/* ioctl's for get/set debug */
#define ETUNIOSDEBUG      _IOW('e', 100, int)
#define ETUNIOSHASFCS	_IOW('e', 101, int)
#define ETUNIOIFATTACH	_IOW('e', 102, struct ifreq *)
#define ETUNIOIFDETACH	_IOR('e', 103, int)

#define ETUNDEBUG	if (etundebug) printf
int	etundebug = 0;
int 	etunhasfcs = 1;

struct etun_softc etunctl[NETUN];
extern int ifqmaxlen;

/* for pdev attach function */
void	etunattach __P((int));

/* for ifnet structure */
int	etunif_ioctl __P((struct ifnet *, u_long, caddr_t));
void	etunif_start __P((struct ifnet *));

/* for cdevsw */
int etunopen __P((dev_t, int, int, struct proc *));
int etunclose __P((dev_t, int, int, struct proc *));
int etunioctl __P((dev_t, u_long, caddr_t, int, struct proc *));
int etunread __P((dev_t, struct uio *, int));
int etunwrite __P((dev_t, struct uio *, int));
int etunpoll __P((dev_t, int, struct proc *));

/* static functions */
static void etuninit __P((struct etun_softc *));
static int etunifattach __P((struct etun_softc *, struct ifreq *));
static int etunifdetach __P((struct etun_softc *));


void
etunattach(unused)
	int unused;
{
	int i;
	struct ifnet *ifp;
	u_int8_t myaddr[ETHER_ADDR_LEN];

	myaddr[0] = myaddr[1] = 0;
	myaddr[3] = 0x12;
	myaddr[4] = 0x34;
	myaddr[5] = 0x56;

	for (i = 0; i < NETUN; i++) {
		myaddr[2] = i;

		ifp = &etunctl[i].etun_ethercom.ec_if;
		etunctl[i].bridgeif = NULL;
		sprintf(ifp->if_xname, "etun%d", i);
		ifp->if_softc = &etunctl[i];
		ifp->if_ioctl = etunif_ioctl;
		ifp->if_start = etunif_start;
		ifp->if_snd.ifq_maxlen = ifqmaxlen;
		ifp->if_flags = IFF_BROADCAST | IFF_NOTRAILERS 
				| IFF_SIMPLEX | IFF_MULTICAST; 
		IFQ_SET_READY(&ifp->if_snd);
		if_attach(ifp);
		ether_ifattach(ifp, myaddr);
	}
	printf("etun: ready\n");
}

/*
 * Process an ioctl request.
 */
int
etunif_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	caddr_t	data;
{
	struct etun_softc *sc = (struct etun_softc *)ifp->if_softc;
	struct ifaddr *ifa = (struct ifaddr *) data;
	struct ifreq *ifr = (struct ifreq *) data;
	int		error = 0, s;

	ETUNDEBUG("%s: etunif_ioctl\n", ifp->if_xname);

	s = splnet();
	switch(cmd) {
	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;
		switch(ifa->ifa_addr->sa_family){
#ifdef INET
		case AF_INET:
			etuninit(sc);
			arp_ifinit(ifp, ifa);
			ETUNDEBUG("%s: address set\n", ifp->if_xname);
			break;
#endif /* INET */
#ifdef NS
	       case AF_NS:
		    {
			 struct ns_addr *ina = &IA_SNS(ifa)->sns_addr;

			 if (ns_nullhost(*ina))
				ina->x_host = *(union ns_host *)
				    LLADDR(ifp->if_sadl);
			 else
				bcopy(ina->x_host.c_host, LLADDR(ifp->if_sadl),
				    ifp->if_addrlen);
			 /* Set new address. */
			 etuninit(sc);
			 break;
		    }
#endif /* NS */
		default:
			etuninit(sc);
		}
		break;

	case SIOCADDMULTI:
	case SIOCDELMULTI: {
		error = (cmd == SIOCADDMULTI) ?
		    ether_addmulti(ifr, &sc->etun_ethercom) :
		    ether_delmulti(ifr, &sc->etun_ethercom);
		break;
	}
	case SIOCSIFFLAGS:
		
		break;
	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}

void
etunif_start(ifp)
	struct ifnet   *ifp;
{
	struct etun_softc *sc = ifp->if_softc;
	struct proc	*p;
	int s;

	ETUNDEBUG("%s: etunif_start\n", ifp->if_xname);

	if ((sc->etun_flags & ETUNF_READY) != ETUNF_READY) {
		ETUNDEBUG ("%s: not ready 0%o\n", ifp->if_xname,
			  sc->etun_flags);
		return;
	}

	s = splnet();

	/* Don't do anything if output is active. */
	if ((ifp->if_flags & IFF_OACTIVE) != 0) {
		splx(s);
		return;
	}

	ifp->if_flags |= IFF_OACTIVE;

	if (sc->etun_flags & ETUNF_RWAIT) {
		sc->etun_flags &= ~ETUNF_RWAIT;
		wakeup((caddr_t)sc);
	}
	if (sc->etun_flags & ETUNF_ASYNC && sc->etun_pgrp) {
		if (sc->etun_pgrp > 0)
			gsignal(sc->etun_pgrp, SIGIO);
		else if ((p = pfind(-sc->etun_pgrp)) != NULL)
			psignal(p, SIGIO);
	}
	selwakeup(&sc->etun_rsel);
	ifp->if_flags &= ~IFF_OACTIVE;
	splx(s);
}

static void
etuninit(sc)
	struct etun_softc *sc;
{
	struct ifnet	*ifp = &sc->etun_ethercom.ec_if;

	ETUNDEBUG("%s: etuninit\n", ifp->if_xname);

	ifp->if_flags |= IFF_UP | IFF_RUNNING;

	return;
}

/* attach ethernet interface to bridge between bif and etun */
static int 
etunifattach(struct etun_softc *sc, struct ifreq *ifr)
{
	struct ifnet *ifp = &sc->etun_ethercom.ec_if;
	struct ifnet *bifp = ifunit(ifr->ifr_name);
#ifdef MAC_REWRITE
	struct ifaddr ifa; 
#endif 
	int error;

	ETUNDEBUG("%s: etunifattach\n", ifp->if_xname);

	if(sc->bridgeif != NULL)
		return EBUSY;
	if(bifp == NULL)
		return ENXIO;
	if(bifp->if_bridge != NULL)
		return EBUSY;

	if(bifp->if_type != IFT_ETHER)
		return EOPNOTSUPP;

#ifdef MAC_REWRITE
	/* save link address of bridge ether i/f */
	memset(&ifa, 0, sizeof(ifa));
	sc->bifaddr = *(bifp->if_sadl);

	/* overwrite etunif's mac addr into bif */
	ifa.ifa_addr = (struct sockaddr *)ifp->if_sadl;
	error = bifp->if_ioctl(bifp, SIOCSIFADDR, (caddr_t)&ifa);
	if(error){
		printf("%s: etunifattach: mac write failed.\n", 
			ifp->if_xname);
		return error;
	}
#endif
	if((error = ifpromisc(ifp, 1)) != 0)
		return error;	
	if((error = ifpromisc(bifp, 1)) != 0){
		ifpromisc(ifp, 0);
		return error;	
	}

	bifp->if_bridge = ifp;
	ifp->if_bridge = bifp;
	sc->bridgeif = bifp;
	return 0;
}

static int 
etunifdetach(struct etun_softc *sc)
{
	struct ifnet *ifp = &sc->etun_ethercom.ec_if;
	struct ifnet *bifp = sc->bridgeif;
#ifdef MAC_REWRITE
	struct ifaddr ifa; 
	int error = 0;
#endif 
	ETUNDEBUG("%s: etunifdetach\n", ifp->if_xname);

	if(sc->bridgeif == NULL)
		return ENXIO;
	sc->bridgeif->if_bridge = NULL; 
	ifp->if_bridge = NULL;
	sc->bridgeif = NULL;

#ifdef MAC_REWRITE
	/* restore mac address of bifp */
	memset(&ifa, 0, sizeof(ifa));
	ifa.ifa_addr = (struct sockaddr *)&sc->bifaddr;
	error = bifp->if_ioctl(bifp, SIOCSIFADDR, (caddr_t)&ifa);
	if(error){
		printf("%s: etunifdetach: mac restore failed.\n", 
			ifp->if_xname);
		return error;
	}
#endif
	ifpromisc(ifp, 0);
	ifpromisc(bifp, 0);
	return 0;
}

/*
 * etunnel open - must be superuser & the device must be
 * configured in
 */
int
etunopen(dev, flag, mode, p)
	dev_t	dev;
	int	flag, mode;
	struct proc *p;
{
	struct ifnet	*ifp;
	struct etun_softc *sc;
	int	unit, error;

	if ((error = suser(p->p_ucred, &p->p_acflag)) != 0)
		return (error);

	if ((unit = minor(dev)) >= NETUN)
		return (ENXIO);

	sc = &etunctl[unit];
	if (sc->etun_flags & ETUNF_OPEN)
		return ENXIO;

	etunclose(dev, flag, mode, p);
	ifp = &sc->etun_ethercom.ec_if;
	sc->etun_flags |= ETUNF_OPEN;
	ETUNDEBUG("%s: open\n", ifp->if_xname);
	return (0);
}

/*
 * etunclose - close the device - mark i/f down 
 */
int
etunclose(dev, flag, mode, p)
	dev_t	dev;
	int	flag;
	int	mode;
	struct proc *p;
{
	int	unit = minor(dev), s;
	struct etun_softc *sc = &etunctl[unit];
	struct ifnet	*ifp = &sc->etun_ethercom.ec_if;
	struct mbuf	*m;

	sc->etun_flags &= ~ETUNF_OPEN;

	/*
	 * junk all pending output
	 */
	do {
		s = splnet();
		IF_DEQUEUE(&ifp->if_snd, m);
		splx(s);
		if (m)
			m_freem(m);
	} while (m);

	if (ifp->if_flags & IFF_UP) {
		s = splnet();
		if_down(ifp);
		splx(s);
	}
	sc->etun_pgrp = 0;
	selwakeup(&sc->etun_rsel);
		
	etunifdetach(sc);
	ETUNDEBUG ("%s: closed\n", ifp->if_xname);
	etundebug = 0;
	return (0);
}

/*
 * the cdevsw interface is now pretty minimal.
 */
int
etunioctl(dev, cmd, data, flag, p)
	dev_t		dev;
	u_long		cmd;
	caddr_t		data;
	int		flag;
	struct proc	*p;
{
	int		unit = minor(dev), s;
	struct etun_softc *sc = &etunctl[unit];
	int error = 0;

	switch (cmd) {
	case ETUNIOSDEBUG:
		etundebug = *(int *)data;
		break;

	case ETUNIOSHASFCS:
		etunhasfcs = *(int *)data;
		break;

	case ETUNIOIFATTACH:
		error = etunifattach(sc, (struct ifreq *)data);
		break;

	case ETUNIOIFDETACH:
		error = etunifdetach(sc);
		break;
		
	case FIONBIO:
		if (*(int *)data)
			sc->etun_flags |= ETUNF_NBIO;
		else
			sc->etun_flags &= ~ETUNF_NBIO;
		break;

	case FIOASYNC:
		if (*(int *)data)
			sc->etun_flags |= ETUNF_ASYNC;
		else
			sc->etun_flags &= ~ETUNF_ASYNC;
		break;

	case FIONREAD:
		s = splnet();
		if (sc->etun_ethercom.ec_if.if_snd.ifq_head)
			*(int *)data = sc->etun_ethercom.ec_if.if_snd.ifq_head->m_pkthdr.len;
		else	
			*(int *)data = 0;
		splx(s);
		break;

	case TIOCSPGRP:
		sc->etun_pgrp = *(int *)data;
		break;

	case TIOCGPGRP:
		*(int *)data = sc->etun_pgrp;
		break;

	default:
		error = ENOTTY;
	}
	return error;
}

/*
 * The cdevsw read interface - reads a packet at a time, or at
 * least as much of a packet as can be read.
 */
int
etunread(dev, uio, ioflag)
	dev_t		dev;
	struct uio	*uio;
	int		ioflag;
{
	int		unit = minor(dev);
	struct etun_softc *sc = &etunctl[unit];
	struct ifnet	*ifp = &sc->etun_ethercom.ec_if;
	struct mbuf	*m, *m0;
	int		error=0, len, s;

	ETUNDEBUG ("%s: read\n", ifp->if_xname);
	if ((sc->etun_flags & ETUNF_READY) != ETUNF_READY) {
		ETUNDEBUG ("%s: not ready 0%o\n", ifp->if_xname, sc->etun_flags);
		return EHOSTDOWN;
	}

	sc->etun_flags &= ~ETUNF_RWAIT;

	s = splnet();
	do {
		IF_DEQUEUE(&ifp->if_snd, m0);
		if (m0 == 0) {
			if (sc->etun_flags & ETUNF_NBIO) {
				splx(s);
				return (EWOULDBLOCK);
			}
			sc->etun_flags |= ETUNF_RWAIT;
			if (tsleep((caddr_t)sc, PZERO|PCATCH, "etunread", 0)) {
				splx(s);
				return (EINTR);
			}
		}
	} while (m0 == 0);
	splx(s);

#if NBPFILTER > 0
	if (ifp->if_bpf) 
		bpf_mtap(ifp->if_bpf, m0);
#endif
	while (m0 && uio->uio_resid > 0 && error == 0) {
		len = min(uio->uio_resid, m0->m_len);
		if (len == 0)
			break;
		error = uiomove(mtod(m0, caddr_t), len, uio);
		MFREE(m0, m);
		m0 = m;
	}

	if (m0) {
		ETUNDEBUG("Dropping mbuf\n");
		m_freem(m0);
	}
	if (error)
		ifp->if_ierrors++;
	return (error);
}

/*
 * the cdevsw write interface - an atomic write is a packet - or else!
 */
int
etunwrite(dev, uio, ioflag)
	dev_t		dev;
	struct uio	*uio;
	int		ioflag;
{
	int		unit = minor (dev);
	struct etun_softc *sc = &etunctl[unit];
	struct ifnet	*ifp = &sc->etun_ethercom.ec_if;
#ifdef USE_M_DEVGET
	char 		buf[ETHER_MAX_LEN+1];
	struct mbuf	*m0;
	int		error=0, tlen;
#else 
	struct mbuf	*m, *m0, *newm;
	int		error=0, tlen, mlen;  
#endif

	ETUNDEBUG("%s: etunwrite\n", ifp->if_xname);

	if (uio->uio_resid < 0){
		ETUNDEBUG("%s: len=%lu!\n", ifp->if_xname,
		    (unsigned long)uio->uio_resid);
		return (EIO);
	}
/*	
	tlen = (uio->uio_resid > ETHER_MAX_LEN ? ETHER_MAX_LEN: uio->uio_resid); */
	tlen = uio->uio_resid;

#ifndef USE_M_DEVGET
	/* get a header mbuf */
	MGETHDR(m0, M_DONTWAIT, MT_DATA);
	if (m0 == NULL)
		return (ENOBUFS);
	m0->m_pkthdr.len = tlen;
	m0->m_pkthdr.rcvif = ifp;
	mlen = MHLEN;

	m = m0;
	while (tlen > 0){
		if(tlen >= MINCLSIZE){
			MCLGET(m, M_DONTWAIT);
			if((m->m_flags & M_EXT) == 0){
				printf("%s:  MCLGET() faild.\n", ifp->if_xname);
				error = ENOBUFS;
				break;
			}
			/* m_adj(m, 2); ether align */

			/* Align the data after the ethernet header */
			m->m_data = (caddr_t) 
			  ALIGN(m->m_data + sizeof(struct ether_header)) 
			    - sizeof(struct ether_header);
			mlen = MCLBYTES;
		}
			
		m->m_len = min(mlen, tlen);
		error = uiomove(mtod (m, caddr_t), m->m_len, uio);
		tlen -= m->m_len;
		if (tlen > 0) {
			MGET (newm, M_DONTWAIT, MT_DATA);
			if (newm == 0) {
				printf("%s:  MGET() faild.\n", ifp->if_xname);
				error = ENOBUFS;
				break;
			}
			mlen = MLEN;
			m->m_next = newm;
			m = newm;
		}
	}
        if (error) {
                if (m0)
                        m_freem (m0);
                ifp->if_ierrors++;
                return (error);
        }

#else /* USE_M_DEVGET */
	error = uiomove(buf, tlen, uio);
	
	if (error) {
		printf("%s: uiomove error.\n", ifp->if_xname);
		ifp->if_ierrors++;
		return (error);
	}

	m0 = m_devget(buf, tlen, 0, ifp, NULL);
	if(m0 == NULL){
		printf("%s: m_devget failed.\n", ifp->if_xname);
		ifp->if_ierrors++;
		return (ENOBUFS);
	}	
#endif /* USE_M_DEVGET */

	ifp->if_ipackets++;
	
	/* strip crc trailer */
	if(etunhasfcs){
		m_adj(m0, -ETHER_CRC_LEN);
		m0->m_flags &= ~M_HASFCS;
	}

	ETUNDEBUG("%s: total len %d \n", ifp->if_xname, m0->m_pkthdr.len);
#if NBPFILTER > 0
	if (ifp->if_bpf) 
		bpf_mtap(ifp->if_bpf, m0);
#endif

	(*ifp->if_input)(ifp, m0);
	return (error);
}

/*
 * etunpoll - the poll interface, this is only useful on reads
 * really. The write detect always returns true, write never blocks
 * anyway, it either accepts the packet or drops it.
 */
int
etunpoll(dev, events, p)
	dev_t		dev;
	int		events;
	struct proc	*p;
{
	int		unit = minor(dev), s;
	struct etun_softc *sc = &etunctl[unit];
	struct ifnet	*ifp = &sc->etun_ethercom.ec_if;
	int		revents = 0;

	s = splnet();
	ETUNDEBUG("%s: etunpoll\n", ifp->if_xname);

	if (events & (POLLIN | POLLRDNORM)) {
		if (ifp->if_snd.ifq_len > 0) {
			ETUNDEBUG("%s: etunpoll q=%d\n", ifp->if_xname,
			    ifp->if_snd.ifq_len);
			revents |= events & (POLLIN | POLLRDNORM);
		} else {
			ETUNDEBUG("%s: etunpoll waiting\n", ifp->if_xname);
			selrecord(p, &sc->etun_rsel);
		}
	}

	if (events & (POLLOUT | POLLWRNORM))
		revents |= events & (POLLOUT | POLLWRNORM);

	splx(s);
	return (revents);
}

#endif  /* NETUN */
